package eknife.erlang;

import edg.edge.Generator;
import edg.graph.*;
import edg.visitor.VoidVisitor;

import java.util.*;

public class ControlFlowGenerator extends VoidVisitor<Void> implements Generator {
	protected static final int INFINITY = -1;

	protected final Set<Node> hangingNodes = new HashSet<>();
	protected final Deque<Set<Node>> returnSet = new LinkedList<>();

	protected final Deque<Set<Node>> clauseHangingStack = new LinkedList<>();
	protected final Deque<Set<Node>> guardHangingStack = new LinkedList<>();
	protected final Deque<Set<Node>> switchStack = new LinkedList<>();
	protected final Deque<Set<Node>> switchSelectorStack = new LinkedList<>();
	protected final Deque<Boolean> switchHasDefaultStack = new LinkedList<>();

	protected boolean absorbentCondition = false;
	protected boolean ifContext = false;

	protected int ttl = INFINITY;
	protected boolean generating = false;
	protected boolean parametersContext = false;
	protected Deque<Node> routineHanging = new LinkedList<>();

	protected final Deque<Set<Node>> lcStack = new LinkedList<>();
	protected final Deque<Set<Node>> prevCaseGuardHanging = new LinkedList<>();

	public ControlFlowGenerator(LAST graph)
	{
		super(graph);
	}

	@Override
	public void generate()
	{
		graph.getRootNode().accept(this, null);
	}

	protected void connectTo(Node n)
	{
		if (!generating)
			return;
		if (ttl > 0 || ttl == INFINITY)
		{
			makeConnections(n);
			if (ttl > 0) ttl--;
		}
		clearHanging();
		hangingNodes.add(n);
	}

	protected void makeConnections(Node n)
	{
		for (Node src : hangingNodes)
			graph.addEdge(src, n, Edge.Type.ControlFlow);
	}

	protected void clearHanging() {
		hangingNodes.clear();
	}

	@Override
	public void visitRoutine(Node n, Void arg)
	{
		assert hangingNodes.isEmpty();
		clauseHangingStack.push(new HashSet<>());
		guardHangingStack.push(new HashSet<>());
		generating = true;
		hangingNodes.add(n);
		routineHanging.push(n);
		super.visitRoutine(n, arg);
		routineHanging.pop();
		hangingNodes.clear();
		hangingNodes.addAll(clauseHangingStack.peek());
		connectTo(n);
		clauseHangingStack.pop();
		guardHangingStack.pop();
		generating = false;
		clearHanging();
	}

	@Override
	public void visitAnonymousRoutine(Node n, Void arg)
	{
		clauseHangingStack.push(new HashSet<>());
		guardHangingStack.push(new HashSet<>());
		routineHanging.push(n);
		connectTo(n);
		super.visitAnonymousRoutine(n, arg);
		routineHanging.pop();
		hangingNodes.clear();
		hangingNodes.addAll(clauseHangingStack.peek());
		connectTo(n);
		guardHangingStack.pop();
		clauseHangingStack.pop();
	}

	@Override
	public void visitClause(Node n, Void arg)
	{
		if (graph.getParent(n).equals(routineHanging.peek()))
		{
			hangingNodes.clear();
			hangingNodes.add(routineHanging.peek());
		}

		connectTo(n);
		clauseHangingStack.peek().addAll(hangingNodes);

		//assert returnSet.isEmpty();
		returnSet.push(new HashSet<>());
		graph.getChild(n, Node.Type.ParameterIn).accept(this, arg);
		graph.getChild(n, Node.Type.Parameters).accept(this, arg);
		graph.getChild(n, Node.Type.Guard).accept(this, arg);
		guardHangingStack.peek().addAll(hangingNodes);
		graph.getChild(n, Node.Type.Body).accept(this, arg);
		hangingNodes.addAll(returnSet.peek());
		returnSet.pop();
		graph.getChild(n, Node.Type.ParameterOut).accept(this, arg);
		hangingNodes.addAll(guardHangingStack.peek());
		guardHangingStack.peek().clear();
		connectTo(n);
	}

	@Override
	public void visitEnclosed(Node n, Void arg)
	{
		super.visitEnclosed(n, arg);
		connectTo(n);
	}

	@Override
	public void visitEquality(Node n, Void arg)
	{
		new EDG(graph).getChild(n, Node.Type.Value).accept(this, arg);
		new EDG(graph).getChild(n, Node.Type.Pattern).accept(this, arg);
		connectTo(n);
	}

	@Override
	public void visitOperation(Node n, Void arg)
	{
		super.visitOperation(n, arg);
		connectTo(n);
	}

	@Override
	public void visitDataConstructor(Node n, Void arg)
	{
		super.visitDataConstructor(n, arg);
		connectTo(n);
	}

	@Override
	public void visitDataConstructorAccess(Node n, Void arg)
	{
		super.visitDataConstructorAccess(n, arg);
		connectTo(n);
	}

	@Override
	public void visitFieldAccess(Node n, Void arg)
	{
		new EDG(graph).getChild(n, Node.Type.Index).accept(this, arg);
		new EDG(graph).getChild(n, Node.Type.Variable).accept(this, arg);
	}

	@Override
	public void visitSwitch(Node n, Void arg)
	{
		if (n.getInfo().getConstruction().equals("if"))
			ifContext = true;
		else
			ifContext = false;

		returnSet.push(new HashSet<>());
		switchStack.push(new HashSet<>());
		graph.getChild(n, Node.Type.Selector).accept(this, arg);
		graph.getChild(n, Node.Type.Cases).accept(this, arg);
		hangingNodes.clear();
		hangingNodes.addAll(returnSet.peek());
		connectTo(n);
		returnSet.pop();

		absorbentCondition = false;
		ifContext = false;
	}

	@Override
	public void visitCases(Node n, Void arg){
		prevCaseGuardHanging.push(new HashSet<>());
		super.visitCases(n, arg);
		returnSet.peek().addAll(Set.copyOf(prevCaseGuardHanging.peek()));
		prevCaseGuardHanging.pop();
	}

	@Override
	public void visitCase(Node n, Void arg)
	{
		if(!absorbentCondition) {
			hangingNodes.addAll(prevCaseGuardHanging.peekFirst());
			prevCaseGuardHanging.peek().clear();
			graph.getChild(n, Node.Type.Selectable).accept(this, arg);
			graph.getChild(n, Node.Type.Guard).accept(this, arg);
			prevCaseGuardHanging.peek().addAll(Set.copyOf(hangingNodes));

			boolean prevIfContext = ifContext;
			boolean prevAbsorbentCondition = absorbentCondition;
			graph.getChild(n, Node.Type.Body).accept(this, arg);
			absorbentCondition = prevAbsorbentCondition;
			ifContext = prevIfContext;

			returnSet.peek().addAll(Set.copyOf(hangingNodes));
		}
	}

	@Override
	public void visitGuard(Node n, Void arg){
		if (this.isIfTrueGuard(n))
			absorbentCondition = true;
		super.visitGuard(n,arg);
	}

	@Override
	public void visitCall(Node n, Void arg)
	{
		graph.getChild(n, Node.Type.Callee).accept(this, arg);
		graph.getChild(n, Node.Type.ArgumentIn).accept(this, arg);
		graph.getChild(n, Node.Type.Arguments).accept(this, arg);
		graph.getChild(n, Node.Type.ArgumentOut).accept(this, arg);
		connectTo(n);
	}

	@Override
	public void visitCallee(Node n, Void arg)
	{
		super.visitCallee(n, arg);
		connectTo(n);
	}

	@Override
	public void visitVariable(Node n, Void arg)
	{
		connectTo(n);
		if (parametersContext &&
				((Variable) n).getContext() == Variable.Context.Use)
			guardHangingStack.peek().addAll(hangingNodes);
	}

	@Override
	public void visitLiteral(Node n, Void arg)
	{
		connectTo(n);
		if (parametersContext)
			guardHangingStack.peek().addAll(hangingNodes);
	}

	@Override
	public void visitReturn(Node n, Void arg)
	{
		super.visitReturn(n, arg);
		connectTo(n);
		returnSet.peek().add(n);
		clearHanging();
	}

	@Override
	public void visitParameterIn(Node n, Void arg)
	{
		connectTo(n);
	}

	@Override
	public void visitParameters(Node n, Void arg)
	{
		parametersContext = true;
		super.visitParameters(n, arg);
		parametersContext = false;
		connectTo(n);
	}

	@Override
	public void visitParameterOut(Node n, Void arg)
	{
		connectTo(n);
	}

	@Override
	public void visitArgumentIn(Node n, Void arg)
	{
		connectTo(n);
	}

	@Override
	public void visitArgumentOut(Node n, Void arg)
	{
		connectTo(n);
	}


	// LIST COMPREHENSIONS AND LISTS IN LIST COMPREHENSIONS
	@Override
	public void visitListComprehension(Node n, Void arg) {
		lcStack.push(new HashSet<>());
		graph.getChild(n, Node.Type.Restrictions).accept(this, arg);
		graph.getChild(n, Node.Type.Value).accept(this, arg);
		ttl = 1;
		hangingNodes.addAll(lcStack.peekFirst());
		Node restrictions = graph.getChild(n, Node.Type.Restrictions);
		Node generator = graph.getChild(restrictions, Node.Type.Generator); // ERROR: More than 1 generator?
		graph.getChild(generator, Node.Type.Iterator).accept(this, arg);
		ttl = INFINITY;
		lcStack.pop();
		connectTo(n);
	}
	@Override
	public void visitRestrictions(Node n, Void arg)
	{
		graph.getChildren(n, Node.Type.Generator).stream()
				.forEach(node -> node.accept(this, arg));
		graph.getChildren(n, Node.Type.Filter).stream()
				.forEach(node -> node.accept(this, arg));
	}

	@Override
	public void visitFilter(Node n, Void arg) {
		super.visitFilter(n,arg);
		lcStack.peekFirst().addAll(hangingNodes);
	}

	@Override
	public void visitGenerator(Node n, Void arg)
	{
		new EDG(graph).getChild(n, Node.Type.Iterator).accept(this, arg);
		new EDG(graph).getChild(n, Node.Type.Variable).accept(this, arg);
		lcStack.peekFirst().addAll(hangingNodes);
	}
	@Override
	public void visitList(Node n, Void arg) {
		super.visitList(n, arg);
		connectTo(n);
	}

	private boolean isIfTrueGuard(Node n) {
		List<Node> guardChildren = graph.getChildren(n);
		if (guardChildren.isEmpty())
			return false;

		Node exprNode = graph.getChildren(
							graph.getChildren(
									guardChildren.get(0)
							).get(0)
						).get(0);
		return exprNode.getType() == Node.Type.Literal &&
				exprNode.getName().equals("true") &&
				ifContext;
	}


}