/*
 * e-Knife, a program slicing tool for Erlang based on the EDG
 * Copyright (c) 2021. David Insa, Sergio Pérez, Josep Silva, Salvador Tamarit.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package eknife.erlang;

import java.io.File;
import java.util.*;

import com.ericsson.otp.erlang.OtpErlangAtom;
import com.ericsson.otp.erlang.OtpErlangList;
import com.ericsson.otp.erlang.OtpErlangLong;
import com.ericsson.otp.erlang.OtpErlangObject;
import com.ericsson.otp.erlang.OtpErlangString;
import com.ericsson.otp.erlang.OtpErlangTuple;

import edg.LDASTNodeInfo;
import edg.graph.EDG;
import edg.graph.Node;
import eknife.config.Config;

public class ErlangCodeFactory
{
	/********************************************************************************************************************************/
	/************************************************************ STATIC ************************************************************/
	/********************************************************************************************************************************/
	private static final Config config = Config.getConfig();

	public static void createErlangFile(File outputFile, EDG edg)
	{
		ErlangCodeFactory.createErlangFile(outputFile, edg, null);
	}
	public static void createErlangFile(File outputFile, EDG edg, Set<Node> slice)
	{
		final ErlangCodeFactory erlangFactory = new ErlangCodeFactory(edg, slice);
		final OtpErlangList outDir = new OtpErlangList(outputFile.toString());
		final OtpErlangList asts = erlangFactory.generate();
		final Launcher launcher = Launcher.getLauncher();
		launcher.launch("saver", "save", outDir, asts);
	}

	/********************************************************************************************************************************/
	/************************************************************ OBJECT ************************************************************/
	/********************************************************************************************************************************/
	private final EDG edg;
	private final Set<Node> slice;
	private final List<String> functions = new LinkedList<String>();
	private final Map<Integer, Boolean> funundef = new Hashtable<Integer, Boolean>();

	private ErlangCodeFactory(EDG edg, Set<Node> slice)
	{
		this.edg = edg;
		this.slice = slice;
	}

	private OtpErlangList generate()
	{
		final Node root = this.edg.getRootNode();
		final List<Node> modules = edg.getChildren(root);
		final List<OtpErlangObject> modules0 = new LinkedList<OtpErlangObject>();

		for (Node module : modules)
		{
			if (this.slice != null && !this.slice.contains(module))
				continue;

			final LDASTNodeInfo ldNodeInfo = module.getInfo();
			final String archive = ldNodeInfo.getFile();
			final OtpErlangTuple moduleTuple = this.parseModule("./" + archive, module);
			modules0.add(moduleTuple);
		}

		final OtpErlangObject[] modulesElements = this.toArray(modules0);
		return new OtpErlangList(modulesElements);
	}

	// Modules
	private OtpErlangTuple parseModule(String file, Node module)
	{
		final OtpErlangObject[] moduleElements = new OtpErlangObject[2];

		this.functions.clear();
		this.funundef.clear();
		moduleElements[0] = new OtpErlangAtom(file);
		moduleElements[1] = this.parseModule(module);

		return new OtpErlangTuple(moduleElements);
	}
	private OtpErlangList parseModule(Node module)
	{
		final List<OtpErlangObject> bodyList = new LinkedList<OtpErlangObject>();

		// Others
		final LDASTNodeInfo ldNodeInfo = module.getInfo();
		final OtpErlangList others = (OtpErlangList) ldNodeInfo.getInfo()[0];
		for (OtpErlangObject other : others)
			bodyList.add(other);

		// Functions
		final List<Node> routines = edg.getChildren(module);
		for (Node routine : routines)
		{
			if (this.slice != null && !this.slice.contains(routine))
				continue;

			final OtpErlangTuple functionTuple = this.parseRoutine(routine);
			bodyList.add(functionTuple);
		}

		// Eof
		final OtpErlangObject[] eofElements = new OtpErlangObject[2];
		eofElements[0] = new OtpErlangAtom("eof");
		eofElements[1] = new OtpErlangLong(1);
		final OtpErlangTuple eofTuple = new OtpErlangTuple(eofElements);
		bodyList.add(eofTuple);

		// Clear ast
		this.removeFile(bodyList);
		this.removeExports(bodyList);

		return new OtpErlangList(this.toArray(bodyList));
	}
	private OtpErlangTuple parseRoutine(Node routine)
	{
		final String routineName = routine.getName();
		final LDASTNodeInfo ldNodeInfo = routine.getInfo();
		final long routineArity = (int) ldNodeInfo.getInfo()[0];
		final List<Node> clauses = edg.getChildrenNonResult(routine);
		final OtpErlangObject[] functionElements = new OtpErlangObject[5];

		functionElements[0] = new OtpErlangAtom("function");
		functionElements[1] = new OtpErlangLong(1);
		functionElements[2] = new OtpErlangAtom(routineName);
		functionElements[3] = new OtpErlangLong(routineArity);
		functionElements[4] = this.parseClauses(clauses);

		this.functions.add(routineName + "/" + routineArity);

		return new OtpErlangTuple(functionElements);
	}
	private OtpErlangList parseClauses(List<Node> clauses)
	{
		final List<OtpErlangObject> clausesList = new LinkedList<OtpErlangObject>();

		for (Node clause : clauses)
		{
			if (this.slice != null && !this.slice.contains(clause))
				continue;

			final OtpErlangTuple clauseTuple = this.parseClause(clause);
			clausesList.add(clauseTuple);
		}
		if (clausesList.isEmpty())
			clausesList.add(this.createEmptyClause());

		final OtpErlangObject[] clausesElements = this.toArray(clausesList);
		return new OtpErlangList(clausesElements);
	}
	private OtpErlangTuple parseClause(Node clause)
	{
		final Node parameters = edg.getChild(clause, Node.Type.Parameters);
		final List<Node> parameterChildren = edg.getChildrenNonResult(parameters);
		final Node guard = edg.getChild(clause, Node.Type.Guard);
		final Node body = edg.getChild(clause, Node.Type.Body);
		final List<Node> expressions = edg.getChildrenNonResult(body);
		final OtpErlangObject[] clauseElements = new OtpErlangObject[5];

		clauseElements[0] = new OtpErlangAtom("clause");
		clauseElements[1] = new OtpErlangLong(1);
		clauseElements[2] = this.parse(parameterChildren, true);
		clauseElements[3] = this.parseGuard(guard);
		OtpErlangList exprs = this.parse(expressions, false);
		if (exprs.elements().length == 0) {
			final OtpErlangObject[] expressionsElements = this.toArray(List.of(this.parseEmptyLiteral()));
			clauseElements[4] = new OtpErlangList(expressionsElements);
		} else
			clauseElements[4] = exprs;

		return new OtpErlangTuple(clauseElements);
	}

	private OtpErlangList parseGuard(Node guard)
	{
		final List<Node> guardChildren = edg.getChildrenNonResult(guard);
		if (guardChildren.isEmpty())
			return new OtpErlangList();

		final List<OtpErlangObject> orGuardsList = new LinkedList<OtpErlangObject>();
		final Node orGuard = guardChildren.get(0);
		final List<Node> andGuards = edg.getChildrenNonResult(orGuard);

		for (Node andGuard : andGuards)
		{
			if (this.slice != null && !this.slice.contains(andGuard))
				continue;

			final List<OtpErlangObject> andGuardList = new LinkedList<OtpErlangObject>();
			final List<Node> andGuardChildren = edg.getChildrenNonResult(andGuard);

			for (Node guardNode : andGuardChildren)
			{
				if (this.slice != null && !this.slice.contains(guardNode))
					continue;

				andGuardList.add(this.parse(guardNode).get(0));
			}

			final OtpErlangObject[] andGuardElements = this.toArray(andGuardList);
			final OtpErlangList andGuardsList = new OtpErlangList(andGuardElements);
			orGuardsList.add(andGuardsList);
		}

		final OtpErlangObject[] orGuardsElements = this.toArray(orGuardsList);
		return new OtpErlangList(orGuardsElements);
	}

	// Expressions
	private OtpErlangList parse(List<Node> nodes, boolean transformUnused)
	{
		final List<OtpErlangObject> expressionsList = new LinkedList<>();

		for (Node node : nodes)
		{
			if (!transformUnused && this.slice != null &&
					!this.slice.contains(node) && !mayContainInternalSliceCode(node))
				continue;

			final List<OtpErlangTuple> expressionTuple = this.parse(node, transformUnused);
			expressionsList.addAll(expressionTuple);
		}
		if (!nodes.isEmpty() && expressionsList.isEmpty() && transformUnused)
			expressionsList.add(this.parseEmptyLiteral());

		final OtpErlangObject[] expressionsElements = this.toArray(expressionsList);
		return new OtpErlangList(expressionsElements);
	}
	private List<OtpErlangTuple> parse(Node node)
	{
		return this.parse(node, true);
	}

	private List<OtpErlangTuple> parse(Node node, boolean transformUnused)
	{
		if (this.slice != null && !this.slice.contains(node) && !mayContainInternalSliceCode(node))
			if (transformUnused)
				return List.of(this.parseEmpty(node));
			else
				return List.of();

		switch (node.getType())
		{
			case Block:
				return List.of(this.parseBlock(node));
			case Switch:
				return this.parseSwitch(node);
			case Call:
				return this.parseCall(node);
			case Variable:
				return List.of(this.parseVariable(node));
			case Literal:
				return List.of(this.parseLiteral(node));
			case Operation:
				return this.parseOperation(node);
			case Equality:
				return this.parseEquality(node);
			case DataConstructor:
				return this.parseDataConstructor(node);
			case List:
				return this.parseList(node);
			case ListComprehension:
				return List.of(this.parseListComprehension(node));
			case Generator:
				return List.of(this.parseGenerator(node));
			case Filter:
				return List.of(this.parseFilter(node));
			case Return:
				return this.parseReturn(node);
			case Routine:
			case AnonymousRoutine:
				return List.of(this.parseAnonymousRoutine(node));
			default:
				throw new RuntimeException("Expression type not contemplated: " + node.getType());
		}
	}

	private OtpErlangTuple parseBlock(Node block)
	{
		final List<Node> blockExpressions = edg.getChildrenNonResult(block);
		final OtpErlangObject[] blockElements = new OtpErlangObject[3];

		blockElements[0] = new OtpErlangAtom("block");
		blockElements[1] = new OtpErlangLong(1);
		blockElements[2] = this.parse(blockExpressions, false);

		return new OtpErlangTuple(blockElements);
	}
	private List<OtpErlangTuple> parseSwitch(Node _switch)
	{
		final LDASTNodeInfo ldNodeInfo = _switch.getInfo();
		final String construction = ldNodeInfo.getConstruction();

		switch (construction)
		{
			case "if":
				return this.parseIf(_switch);
			case "case":
				return this.parseCase(_switch);
			default:
				throw new RuntimeException("Switch type not contemplated: " + construction);
		}
	}
	private List<OtpErlangTuple> parseIf(Node _switch)
	{
		final Node cases = edg.getChild(_switch, Node.Type.Cases);

		if (slice != null && !slice.contains(cases))
			return List.of();

		final List<Node> casesChildren = edg.getChildrenNonResult(cases);
		final OtpErlangObject[] ifElements = new OtpErlangObject[3];

		ifElements[0] = new OtpErlangAtom("if");
		ifElements[1] = new OtpErlangLong(1);
		ifElements[2] = this.parseCaseClauses(casesChildren);

		return List.of(new OtpErlangTuple(ifElements));
	}

	private List<OtpErlangTuple> parseCase(Node _switch)
	{
		// Case 1: Only assignments in the selector node (Cases Node is not part of the slice)
		final Node selectorNode = edg.getChild(_switch, Node.Type.Selector);
		final Node selector = edg.getChild(selectorNode, Node.Type.Value);
		final Node cases = edg.getChild(_switch, Node.Type.Cases);

		if (slice != null && !slice.contains(cases))
			return this.parse(selector);

		// Case 2: Cases Node is in the slice, so we build a whole case structure
		final List<Node> casesChildren = edg.getChildrenNonResult(cases);
		final OtpErlangObject[] caseElements = new OtpErlangObject[4];

		caseElements[0] = new OtpErlangAtom("case");
		caseElements[1] = new OtpErlangLong(1);
		caseElements[2] = this.parse(selector).get(0);
		caseElements[3] = this.parseCaseClauses(casesChildren);

		return List.of(new OtpErlangTuple(caseElements));
	}

	private OtpErlangList parseCaseClauses(List<Node> clauses)
	{
		final List<OtpErlangObject> clausesList = new LinkedList<>();

		for (Node clause : clauses)
		{
			if (this.slice != null && !this.slice.contains(clause))
				continue;

			final OtpErlangTuple clauseTuple = this.parseCaseClause(clause);
			clausesList.add(clauseTuple);
		}
		if (clausesList.isEmpty())
			clausesList.add(this.createEmptyClause());

		final OtpErlangObject[] clausesElements = this.toArray(clausesList);
		return new OtpErlangList(clausesElements);
	}

	private OtpErlangTuple parseCaseClause(Node clause)
	{
		final Node selectableNode = edg.getChild(clause, Node.Type.Selectable);
		final List<Node> selectables = edg.getChildrenNonResult(selectableNode);
		final Node guard = edg.getChild(clause, Node.Type.Guard);
		final Node body = edg.getChild(clause, Node.Type.Body);
		final List<Node> expressions = edg.getChildrenNonResult(body);
		final OtpErlangObject[] clauseElements = new OtpErlangObject[5];

		clauseElements[0] = new OtpErlangAtom("clause");
		clauseElements[1] = new OtpErlangLong(1);
		clauseElements[2] = this.parse(selectables, true);
		clauseElements[3] = this.parseGuard(guard);
		OtpErlangList exprs = this.parse(expressions, false);
		if (exprs.elements().length == 0) {
			final OtpErlangObject[] expressionsElements = this.toArray(List.of(this.parseEmptyLiteral()));
			clauseElements[4] = new OtpErlangList(expressionsElements);
		} else
			clauseElements[4] = exprs;

		return new OtpErlangTuple(clauseElements);
	}

	private List<OtpErlangTuple> parseCall(Node call)
	{
		final Node argumentsNode = edg.getChild(call, Node.Type.Arguments);
		final List<Node> arguments = edg.getChildrenNonResult(argumentsNode);

		if (slice != null && !slice.contains(call)){ // TODO: Think a better way
			final List<OtpErlangTuple> containedArgs = new LinkedList<>();
			OtpErlangList parsedArgs = this.parse(arguments, false);
			for (OtpErlangObject parsedArg : parsedArgs) {
				if (!(parsedArg instanceof OtpErlangTuple))
					throw new RuntimeException("The parsed args are not OtpTuples");
				containedArgs.add((OtpErlangTuple) parsedArg);
			}
			return containedArgs;
		}

		final Node callee = edg.getChild(call, Node.Type.Callee);
		final Node scopeNode = edg.getChild(callee, Node.Type.Scope);
		final Node nameNode = edg.getChild(callee, Node.Type.Name);
		final List<Node> scopeChildren = edg.getChildrenNonResult(scopeNode);
		final OtpErlangObject[] callElements = new OtpErlangObject[4];

		callElements[0] = new OtpErlangAtom("call");
		callElements[1] = new OtpErlangLong(1);
		if (!scopeChildren.isEmpty() && (this.slice == null || this.slice.contains(callee)))
		{
			final Node scope = edg.getChild(scopeNode, Node.Type.Value);
			final Node name = edg.getChild(nameNode, Node.Type.Value);
			final OtpErlangObject[] remoteElements = new OtpErlangObject[4];

			remoteElements[0] = new OtpErlangAtom("remote");
			remoteElements[1] = new OtpErlangLong(1);
			remoteElements[2] = this.parse(scope).get(0);
			remoteElements[3] = this.parse(name).get(0);
			callElements[2] = new OtpErlangTuple(remoteElements);
		}
		else
		{
			final Node name = edg.getChild(nameNode, Node.Type.Value);
			callElements[2] = this.parse(name).get(0);
		}
		callElements[3] = this.parse(arguments, true);

		return List.of(new OtpErlangTuple(callElements));
	}
	private List<OtpErlangTuple> parseOperation(Node operation)
	{
		final List<Node> operationExpressions = edg.getChildrenNonResult(operation);

		switch (operationExpressions.size())
		{
			case 1:
				return this.parseUnaryOperation(operation);
			case 2:
				return this.parseBinaryOperation(operation);
			default:
				throw new RuntimeException("Operation size not contemplated: " + operationExpressions.size());
		}
	}
	private List<OtpErlangTuple> parseUnaryOperation(Node operation)
	{
		final String nodeText = operation.getName();
		final Node operationExpression = edg.getChild(operation, 0);
		final OtpErlangObject[] operationElements = new OtpErlangObject[4];

		operationElements[0] = new OtpErlangAtom("op");
		operationElements[1] = new OtpErlangLong(1);
		operationElements[2] = new OtpErlangAtom(nodeText);
		operationElements[3] = this.parse(operationExpression).get(0);

		return List.of(new OtpErlangTuple(operationElements));
	}
	private List<OtpErlangTuple> parseBinaryOperation(Node operation)
	{
		final String nodeText = operation.getName();
		final Node operationExpression1 = edg.getChild(operation, 0);
		final Node operationExpression2 = edg.getChild(operation, 1);

		if (slice != null && !slice.contains(operation)) {
			List<OtpErlangTuple> leftExpr = new LinkedList<>();
			leftExpr.addAll(this.parse(operationExpression1, false));
			List<OtpErlangTuple> rightExpr = this.parse(operationExpression2, false);
			leftExpr.addAll(rightExpr);
			return leftExpr;
		}

		final OtpErlangObject[] operationElements = new OtpErlangObject[5];

		operationElements[0] = new OtpErlangAtom("op");
		operationElements[1] = new OtpErlangLong(1);
		operationElements[2] = new OtpErlangAtom(nodeText);
		operationElements[3] = this.parse(operationExpression1).get(0);
		operationElements[4] = this.parse(operationExpression2).get(0);

		return List.of(new OtpErlangTuple(operationElements));
	}
	private List<OtpErlangTuple> parseDataConstructor(Node dataConstructor)
	{
		final List<Node> dataConstructorExpressions = edg.getChildrenNonResult(dataConstructor);
		if (slice != null && !slice.contains(dataConstructor)) {
			final List<OtpErlangTuple> containedArgs = new LinkedList<>();
			OtpErlangList parsedElems = this.parse(dataConstructorExpressions, false);
			for (OtpErlangObject parsedElem : parsedElems) {
				if (!(parsedElem instanceof OtpErlangTuple))
					throw new RuntimeException("The parsed element are not OtpTuples");
				containedArgs.add((OtpErlangTuple) parsedElem);
			}

			if (containedArgs.isEmpty() && edg.isPatternZone(dataConstructor))
				containedArgs.add(parseEmpty(dataConstructor));

			return containedArgs;
		}

		final OtpErlangObject[] tupleElements = new OtpErlangObject[3];
		tupleElements[0] = new OtpErlangAtom("tuple");
		tupleElements[1] = new OtpErlangLong(1);
		tupleElements[2] = this.parse(dataConstructorExpressions, true);

		return List.of(new OtpErlangTuple(tupleElements));
	}
	private List<OtpErlangTuple> parseList(Node list)
	{
		if (edg.getChildren(list).isEmpty())
		{
			final OtpErlangObject[] nilElements = new OtpErlangObject[2];
			nilElements[0] = new OtpErlangAtom("nil");
			nilElements[1] = new OtpErlangLong(1);
			return List.of(new OtpErlangTuple(nilElements));
		}

		final List<Node> listChildren = edg.getChildrenNonResult(list);
		if (slice != null && !slice.contains(list)) {
			final List<OtpErlangTuple> containedArgs = new LinkedList<>();
			OtpErlangList parsedElems = this.parse(listChildren, false);
			for (OtpErlangObject parsedElem : parsedElems) {
				if (!(parsedElem instanceof OtpErlangTuple))
					throw new RuntimeException("The parsed element are not OtpTuples");
				containedArgs.add((OtpErlangTuple) parsedElem);
			}
			return containedArgs;
		}

		final Node listHead = edg.getChild(list, Node.Type.Value);
		final Node listTail = edg.getChild(list, Node.Type.List);
		final OtpErlangObject[] listElements = new OtpErlangObject[4];

		listElements[0] = new OtpErlangAtom("cons");
		listElements[1] = new OtpErlangLong(1);
		listElements[2] = this.parse(listHead).get(0);
		listElements[3] = this.parse(listTail).get(0);

		return List.of(new OtpErlangTuple(listElements));
	}
	private List<OtpErlangTuple> parseEquality(Node equality)
	{
		final Node equalityPattern = edg.getChild(equality, Node.Type.Pattern);
		final Node equalityValue = edg.getChild(equality, Node.Type.Value);

		if (slice != null && !slice.contains(equalityPattern)) {
			List<OtpErlangTuple> returnValue = this.parse(equalityValue, false);
			return returnValue;
		}

		final OtpErlangObject[] patternMatchingElements = new OtpErlangObject[4];

		patternMatchingElements[0] = new OtpErlangAtom("match");
		patternMatchingElements[1] = new OtpErlangLong(1);
		patternMatchingElements[2] = this.parse(equalityPattern).get(0);
		patternMatchingElements[3] = this.parse(equalityValue).get((0));

		return List.of(new OtpErlangTuple(patternMatchingElements));
	}
	private OtpErlangTuple parseListComprehension(Node listComprehension)
	{
		final List<Node> generators = edg.getChildren(listComprehension, Node.Type.Generator);
		final Node valueNode = edg.getChild(listComprehension, Node.Type.Value);
		final Node value = edg.getChild(valueNode, Node.Type.Value);
		final OtpErlangObject[] listComprehensionElements = new OtpErlangObject[4];

		listComprehensionElements[0] = new OtpErlangAtom("lc");
		listComprehensionElements[1] = new OtpErlangLong(1);
		listComprehensionElements[2] = this.parse(value).get(0);
		listComprehensionElements[3] = this.parse(generators, false);

		return new OtpErlangTuple(listComprehensionElements);
	}
	private OtpErlangTuple parseGenerator(Node generator)
	{
		final Node pattern = edg.getChild(generator, Node.Type.Variable);
		final Node expression = edg.getChild(generator, Node.Type.Iterator);
		final OtpErlangObject[] generatorElements = new OtpErlangObject[4];

		generatorElements[0] = new OtpErlangAtom("generate");
		generatorElements[1] = new OtpErlangLong(1);
		generatorElements[2] = this.parse(pattern).get(0);
		generatorElements[3] = this.parse(expression).get(0);

		return new OtpErlangTuple(generatorElements);
	}
	private OtpErlangTuple parseFilter(Node filter)
	{
		final Node filterExpr = edg.getChild(filter, Node.Type.Value);
		return this.parse(filterExpr).get(0);
	}
	private List<OtpErlangTuple> parseReturn(Node returnNode)
	{
		final Node returnExpression = edg.getChild(returnNode, Node.Type.Value);
		return this.parse(returnExpression);
	}
	private OtpErlangTuple parseAnonymousRoutine(Node anonymousRoutine)
	{
		final String routineName = anonymousRoutine.getName();

		if (routineName == null)
			return this.parseNormalAnonymousRoutine(anonymousRoutine);
		return this.parseNamedAnonymousRoutine(anonymousRoutine);
	}
	private OtpErlangTuple parseNormalAnonymousRoutine(Node anonymousRoutine)
	{
		final LDASTNodeInfo ldNodeInfo = anonymousRoutine.getInfo();
		final boolean sugar = (boolean) ldNodeInfo.getInfo()[0];

		if (!sugar)
		{
			final List<Node> clauses = edg.getChildrenNonResult(anonymousRoutine);
			final OtpErlangObject[] funElements = new OtpErlangObject[3];
			final OtpErlangObject[] clausesElements = new OtpErlangObject[2];

			clausesElements[0] = new OtpErlangAtom("clauses");
			clausesElements[1] = this.parseClauses(clauses);
			funElements[0] = new OtpErlangAtom("fun");
			funElements[1] = new OtpErlangLong(1);
			funElements[2] = new OtpErlangTuple(clausesElements);

			return new OtpErlangTuple(funElements);
		}
		else
		{
			final List<Node> clauses = edg.getChildrenNonResult(anonymousRoutine);
			assert !clauses.isEmpty();
			final Node clause = clauses.get(0);
			final Node body = edg.getChild(clause, Node.Type.Body);
			final Node returnNode = edg.getChildren(body, Node.Type.Return).get(0);
			final Node call = edg.getChild(returnNode, Node.Type.Value);
			final Node callee = edg.getChild(call, Node.Type.Callee);
			final Node scopeNode = edg.getChild(callee, Node.Type.Scope);
			final Node scope = edg.getChildren(scopeNode).isEmpty() ? null :
					edg.getChild(scopeNode, Node.Type.Value);
			final Node nameNode = edg.getChild(callee, Node.Type.Name);
			final Node name = edg.getChildrenNonResult(nameNode).isEmpty() ? null :
					edg.getChild(nameNode, Node.Type.Value);
			final Node argumentsNode = edg.getChild(call, Node.Type.Arguments);
			final List<Node> arguments = edg.getChildrenNonResult(argumentsNode);
			final int arity = arguments.size();

			final OtpErlangObject[] funElements = new OtpErlangObject[3];
			final OtpErlangObject[] functionElements;

			if (scope == null)
			{
				functionElements = new OtpErlangObject[3];
				functionElements[0] = new OtpErlangAtom("function");
				functionElements[1] = new OtpErlangAtom(name.getName());
				functionElements[2] = new OtpErlangLong(arity);
			}
			else
			{
				final OtpErlangObject[] moduleElements = new OtpErlangObject[3];
				final OtpErlangObject[] nameElements = new OtpErlangObject[3];
				final OtpErlangObject[] arityElements = new OtpErlangObject[3];
				final Node.Type scopeType = scope.getType();
				final Node.Type nameType = name.getType();

				moduleElements[0] = new OtpErlangAtom(scopeType == Node.Type.Literal ? "atom" : "var");
				moduleElements[1] = new OtpErlangLong(1);
				moduleElements[2] = new OtpErlangAtom(scope.getName());
				nameElements[0] = new OtpErlangAtom(nameType == Node.Type.Literal ? "atom" : "var");
				nameElements[1] = new OtpErlangLong(1);
				nameElements[2] = new OtpErlangAtom(name.getName());
				arityElements[0] = new OtpErlangAtom("integer");
				arityElements[1] = new OtpErlangLong(1);
				arityElements[2] = new OtpErlangLong(arity);
				functionElements = new OtpErlangObject[4];
				functionElements[0] = new OtpErlangAtom("function");
				functionElements[1] = new OtpErlangTuple(moduleElements);
				functionElements[2] = new OtpErlangTuple(nameElements);
				functionElements[3] = new OtpErlangTuple(arityElements);
			}
			funElements[0] = new OtpErlangAtom("fun");
			funElements[1] = new OtpErlangLong(1);
			funElements[2] = new OtpErlangTuple(functionElements);

			return new OtpErlangTuple(funElements);
		}
	}
	private OtpErlangTuple parseNamedAnonymousRoutine(Node anonymousRoutine)
	{
		final String routineName = anonymousRoutine.getName();
		final List<Node> clauses = edg.getChildrenNonResult(anonymousRoutine);
		final OtpErlangObject[] functionElements = new OtpErlangObject[4];
		final OtpErlangObject[] clausesElements = new OtpErlangObject[2];

		clausesElements[0] = new OtpErlangAtom("clauses");
		clausesElements[1] = this.parseClauses(clauses);
		functionElements[0] = new OtpErlangAtom("named_fun");
		functionElements[1] = new OtpErlangLong(1);
		functionElements[2] = new OtpErlangAtom(routineName);
		functionElements[3] = this.parseClauses(clauses);

		return new OtpErlangTuple(functionElements);
	}

	// Variables & literals
	private OtpErlangTuple parseVariable(Node variable)
	{
		final String variableName = variable.getName();
		final OtpErlangObject[] variableElements = new OtpErlangObject[3];

		variableElements[0] = new OtpErlangAtom("var");
		variableElements[1] = new OtpErlangLong(1);
		variableElements[2] = new OtpErlangAtom(variableName);

		return new OtpErlangTuple(variableElements);
	}
	private OtpErlangTuple parseLiteral(Node literal)
	{
		final LDASTNodeInfo ldNodeInfo = literal.getInfo();
		final String construction = ldNodeInfo.getConstruction();

		switch (construction)
		{
			case "atom":
				return this.parseAtom(literal);
			case "string":
				return this.parseString(literal);
			case "integer":
				return this.parseInteger(literal);
			case "char":
				return this.parseChar(literal);
			default:
				throw new RuntimeException("Literal type not contemplated: " + construction);
		}
	}
	private OtpErlangTuple parseAtom(Node atom)
	{
		final String atomText = atom.getName();
		final OtpErlangObject[] atomElements = new OtpErlangObject[3];

		atomElements[0] = new OtpErlangAtom("atom");
		atomElements[1] = new OtpErlangLong(1);
		atomElements[2] = new OtpErlangAtom(atomText);

		return new OtpErlangTuple(atomElements);
	}
	private OtpErlangTuple parseString(Node string)
	{
		final String stringText = string.getName();
		final OtpErlangObject[] stringElements = new OtpErlangObject[3];

		stringElements[0] = new OtpErlangAtom("string");
		stringElements[1] = new OtpErlangLong(1);
		stringElements[2] = new OtpErlangString(stringText);

		return new OtpErlangTuple(stringElements);
	}
	private OtpErlangTuple parseInteger(Node integer)
	{
		final String nodeText = integer.getName();
		final long value = Long.parseLong(nodeText);
		final OtpErlangObject[] integerElements = new OtpErlangObject[3];

		integerElements[0] = new OtpErlangAtom("integer");
		integerElements[1] = new OtpErlangLong(1);
		integerElements[2] = new OtpErlangLong(value);

		return new OtpErlangTuple(integerElements);
	}
	private OtpErlangTuple parseChar(Node _char)
	{
		final String nodeText = _char.getName();
		final long value = Long.parseLong(nodeText);
		final OtpErlangObject[] charElements = new OtpErlangObject[3];

		charElements[0] = new OtpErlangAtom("char");
		charElements[1] = new OtpErlangLong(1);
		charElements[2] = new OtpErlangLong(value);

		return new OtpErlangTuple(charElements);
	}

	// Auxiliaries
	private OtpErlangTuple parseEmpty(Node node)
	{
		if (edg.isPatternZone(node))
			return this.parseEmptyVar();
		return this.parseEmptyLiteral();
	}
	private OtpErlangTuple parseEmptyVar()
	{
		final OtpErlangObject[] varElements = new OtpErlangObject[3];

		varElements[0] = new OtpErlangAtom("var");
		varElements[1] = new OtpErlangLong(1);
		varElements[2] = new OtpErlangAtom("_");

		return new OtpErlangTuple(varElements);
	}
	private OtpErlangTuple parseEmptyLiteral()
	{
		final OtpErlangObject[] varElements = new OtpErlangObject[3];

		varElements[0] = new OtpErlangAtom("atom");
		varElements[1] = new OtpErlangLong(1);
		varElements[2] = new OtpErlangAtom("sliced");

		return new OtpErlangTuple(varElements);
	}
	private OtpErlangTuple createEmptyClause()
	{
		final OtpErlangObject[] varElements = new OtpErlangObject[5];

		varElements[0] = new OtpErlangAtom("clause");
		varElements[1] = new OtpErlangLong(1);
		varElements[2] = new OtpErlangList(this.parseEmptyVar());
		varElements[3] = new OtpErlangList(new OtpErlangObject[0]);
		varElements[4] = new OtpErlangList(this.parseEmptyLiteral());

		return new OtpErlangTuple(varElements);
	}

	private void removeFile(List<OtpErlangObject> bodyList)
	{
		for (int bodyElementIndex = 0; bodyElementIndex < bodyList.size(); bodyElementIndex++)
		{
			final OtpErlangTuple element = (OtpErlangTuple) bodyList.get(bodyElementIndex);
			final OtpErlangAtom type = (OtpErlangAtom) element.elementAt(2);
			if (type == null || !type.toString().equals("file"))
				continue;

			bodyList.remove(bodyElementIndex);
			break;
		}
	}
	private void removeExports(List<OtpErlangObject> bodyList)
	{
		for (int bodyElementIndex = 0; bodyElementIndex < bodyList.size(); bodyElementIndex++)
		{
			final OtpErlangTuple element = (OtpErlangTuple) bodyList.get(bodyElementIndex);
			final OtpErlangAtom type = (OtpErlangAtom) element.elementAt(2);
			if (type == null || !type.toString().equals("export"))
				continue;

			final OtpErlangList exports = (OtpErlangList) element.elementAt(3);
			final List<OtpErlangObject> newExports = new LinkedList<OtpErlangObject>();

			for (OtpErlangObject export : exports)
			{
				final OtpErlangTuple export0 = (OtpErlangTuple) export;
				final String name = export0.elementAt(0).toString();
				final String arity = export0.elementAt(1).toString();

				if (this.functions.contains(name + "/" + arity))
					newExports.add(export);
			}

			bodyList.remove(bodyElementIndex);
			if (!newExports.isEmpty())
			{
				final OtpErlangObject[] elements = new OtpErlangObject[4];
				elements[0] = new OtpErlangAtom("attribute");
				elements[1] = new OtpErlangLong(1);
				elements[2] = new OtpErlangAtom("export");
				elements[3] = new OtpErlangList(this.toArray(newExports));

				final OtpErlangTuple newElement = new OtpErlangTuple(elements);
				bodyList.add(bodyElementIndex, newElement);
			}
		}
	}
	private OtpErlangObject[] toArray(List<OtpErlangObject> list)
	{
		final int listSize = list.size();
		final OtpErlangObject[] elements = new OtpErlangObject[listSize];

		for (int elementIndex = 0; elementIndex < listSize; elementIndex++)
			elements[elementIndex] = list.get(elementIndex);

		return elements;
	}

	private boolean mayContainInternalSliceCode(Node node)
	{
		switch (node.getType())
		{
			case Operation:
			case Equality:
			case Call:
			case Condition:
			case If:
			case Switch:
			case Selector:
			case Cases:
			case Case:
			case Enclosed:
			case Generator:
			case Iterator:
			case DataConstructor:
			case DataConstructorAccess:
			case List:
			case Return:
				return true;
			default:
				return false;
		}
	}
}